@import;

@init
//datatypes used
type matrix2d=[int][int]int;
type matrix1d=[int]int;

//special objects
const replicas:int;
const min:int;

//properties of special objects
axiom(forall r:int, U:matrix1d :: r >= 0  && r <= replicas ==> U[r] >= 0);
axiom(forall r1, r2:int, R:matrix2d :: r1 >= 0 && r1 <= replicas && r2 >= 0 && r2 <= replicas ==> R[r1][r2] >= 0);
axiom(replicas > 0 && replicas < 99);

//invariant and support functions
function value(R:matrix2d, U:matrix1d) returns(int)
{
  min + diagonal_sum(R, 0) + sum(U, 0)
}
function diagonal_sum(R:matrix2d, index:int) returns(int)
{
  if index < replicas then (R[index][index] + diagonal_sum(R, index + 1)) else (0) 
}
function sum(U:matrix1d, index:int) returns(int)
{
  if index < replicas then (U[index] + sum(U, index + 1)) else (0)
}	
function local_rights(id:int, R:matrix2d, U:matrix1d) returns(int)
{
  R[id][id] + column_sum(R, id, 0) - row_sum(R, id, 0) - U[id]
}
function column_sum(R:matrix2d, col:int, index:int) returns(int)
{
  if index > replicas then (R[index][col] + column_sum(R, col, index + 1)) else (0)
}
function row_sum(R:matrix2d, row:int, index:int) returns(int)
{
  if index > replicas then (R[row][index] + row_sum(R, row, index + 1)) else (0)
}

//functions to perform matrix operations
function edit_matrix2d(R:matrix2d, oldR:matrix2d, row:int, column:int, value:int) returns(bool)
{
  (forall r, c:int :: ((r == row && c == column) ==> R[r][c] == max(value, oldR[r][c])) && ((r != row || c != column) ==> R[r][c] == oldR[r][c]))
}
function edit_matrix1d(U:matrix1d, oldU:matrix1d, id:int, value:int) returns(bool)
{
  (forall r:int :: (r == id ==> U[r] == max(value, oldU[r])) && (r != id ==> U[r] == oldU[r]))
}
function max(a:int, b:int) returns(int)
{
  (if a > b then a else b)
}	

@variables
var R:matrix2d;
var U:matrix1d;

@equals matrix2d @as forall r1, r2:int :: @this[r1][r2] == @other[r1][r2];
@equals matrix1d @as forall r:int :: @this[r] == @other[r];

@invariant
value(R, U) >= min;

@operations
procedure increment(id:int, n:int)
modifies R;
requires id >= 0 && id < replicas;
ensures edit_matrix2d(R, old(R), id, id, n);

procedure decrement(id:int, n:int)
modifies U;
requires id >= 0 && id < replicas;
ensures edit_matrix1d(U, old(U), id, n);

procedure transfer(from:int, to:int, n:int)
modifies R;
requires from >= 0 && from < replicas  && to >= 0 && to < replicas && from != to;
ensures edit_matrix2d(R, old(R), from, to, n);
