@import;

@init
//datatypes used
type matrix2d=[int,int]int;
type matrix1d=[int]int;

//special objects used
const replicas:int;
const min:int;

//properties
axiom(forall r:int, U:matrix1d :: r >= 0  && r <= replicas ==> U[r] >= 0);
axiom(forall r1, r2:int, R:matrix2d :: r1 >= 0 && r1 <= replicas && r2 >= 0 && r2 <= replicas ==> R[r1,r2] >= 0);
axiom(replicas > 0 && replicas < 99);

//support functions
function value(R:matrix2d, U:matrix1d) returns(int)
{
  min + diagonal_sum(R, 0) + sum(U, 0)
}
function diagonal_sum(R:matrix2d, index:int) returns(int)
{
  if index < replicas then (R[index,index] + diagonal_sum(R, index + 1)) else (0) 
}
function sum(U:matrix1d, index:int) returns(int)
{
  if index < replicas then (U[index] + sum(U, index + 1)) else (0)
}	
function local_rights(id:int, R:matrix2d, U:matrix1d) returns(int)
{
  R[id,id] + column_sum(R, id, 0) - row_sum(R, id, 0) - U[id]
}
function column_sum(R:matrix2d, col:int, index:int) returns(int)
{
  if index > replicas then (R[index,col] + column_sum(R, col, index + 1)) else (0)
}
function row_sum(R:matrix2d, row:int, index:int) returns(int)
{
  if index > replicas then (R[row,index] + row_sum(R, row, index + 1)) else (0)
}

function increment_effect(R:matrix2d, oldR:matrix2d, id:int, n:int) returns(bool)
{
  (forall r1, r2:int :: ((r1 == id && r2 == id) ==> R[r1,r2] == oldR[r1,r2] + n) && ((r1 != id || r2 != id) ==> R[r1,r2] == oldR[r1,r2]))
}
function decrement_effect(U:matrix1d, oldU:matrix1d, id:int, n:int) returns(bool)
{
  (forall r:int :: (r == id ==> U[r] == oldU[r] + n) && (r != id ==> U[r] == oldU[r]))
}
function transfer_effect(R:matrix2d, oldR:matrix2d, from:int, to:int, n:int) returns(bool)
{
  (forall r1, r2:int :: ((r1 == from && r2 == to) ==> R[r1,r2] == oldR[r1,r2] + n) && ((r1 != from || r2 != to) ==> R[r1,r2] == oldR[r1,r2]))
}

@variables
var R:matrix2d;
var U:matrix1d;

@equals matrix2d @as forall r1, r2:int :: @this[r1,r2] == @other[r1,r2];
@equals matrix1d @as forall r:int :: @this[r] == @other[r];

@invariant
value(R, U) >= min;

@operations

procedure increment(id:int, n:int)
modifies R;
requires n > 0 && id >= 0 && id < replicas;
ensures increment_effect(R, old(R), id, n);

procedure decrement(id:int, n:int)
modifies U;
requires n > 0  && id >= 0 && id < replicas && local_rights(id, R, U) >= n;
ensures decrement_effect(U, old(U), id, n);

procedure transfer(from:int, to:int, n:int)
modifies R;
requires n > 0  && from >= 0 && from < replicas  && to >= 0 && to < replicas && from != to && local_rights(from, R, U) >= n;
ensures transfer_effect(R, old(R), from, to, n);
