@import;

@init
type Credit = int;
type ReplicaID;


@variables
var localCredit:[ReplicaID]Credit;
var globalCredit:Credit;
var l:ReplicaID;

@equals [ReplicaID]Credit @as forall r:ReplicaID :: @this[r] == @other[r];
@equals Credit @as @this == @other;

@invariant
(forall r:ReplicaID :: localCredit[r] >= 0) && (globalCredit >= 0);
@operations

procedure acquireCredit(replica:ReplicaID, k:Credit)
modifies globalCredit, localCredit;
requires k > 0 && globalCredit >= k && l == replica;
ensures (forall r:ReplicaID :: (r == replica ==> localCredit[r] == old(localCredit)[r] + k) && (r != replica ==> localCredit[r] == old(localCredit)[r])) && globalCredit == old(globalCredit) - k;

procedure releaseCredit(replica:ReplicaID, k:Credit)
modifies globalCredit, localCredit;
requires k > 0 && localCredit[replica] >= k;
ensures (forall r:ReplicaID :: (r == replica ==> localCredit[r] == old(localCredit)[r] - k) && (r != replica ==> localCredit[r] == old(localCredit)[r])) && globalCredit == old(globalCredit) + k;

procedure increment(replica:ReplicaID, k:Credit)
modifies localCredit;
requires k > 0;
ensures forall r:ReplicaID :: (r == replica ==> localCredit[r] == old(localCredit)[r] + k) && (r != replica ==> localCredit[r] == old(localCredit)[r]);

procedure decrement(replica:ReplicaID, k:Credit)
modifies localCredit;
requires k > 0 && localCredit[replica] >= k;
ensures forall r:ReplicaID :: (r == replica ==> localCredit[r] == old(localCredit)[r] - k) && (r != replica ==> localCredit[r] == old(localCredit)[r]);
